#ifndef __RT_CONSUMER_H__
#define __RT_CONSUMER_H__

#include <cstddef>
#ifndef _MSC_VER
#include <inttypes.h>
#else
#if _MSC_VER >= 1600
#include <cstdint>
#else
typedef unsigned __int64 uint64_t;
#endif  //  _MSC_VER >= 1600
#endif // !_MSC_VER
#include <vector>
#include <fire/private/prefix.h>
#include "fire/detail/RT_Feature.h"

namespace FIRE_NAMESPACE
{

class RT_API RT_Consumer
{
  public:
    typedef void (*CtrlCB)(bool is_connected, const class RT_Message *msg,
                                              void *closure);
    typedef void (*MesgCB)(int tag, const class RT_Message *msg,
                                    const class RT_Record   *rec, void *closure);
    typedef void (*NtfyFxn)(void *closure);

    RT_Consumer();
    ~RT_Consumer();

    void set_connection_ctrl_cb(RT_Consumer::CtrlCB cb, void *closure,
                                void **old_closure = NULL);

    // get_connection pointer is valid until next call, or object destruction
    const char *get_connection_info(size_t *dest_sz = NULL) const;
    rt_errno    set_connection_info(const char *connection_string_uri);
    rt_errno    set_connection_info(const char *host_port,
                                    const char *user, const char *passwd);
    rt_errno    set_connection_info(const char *host_port, const char *user,
                                    const char *device_id, const char *key,
                                    const char *counter, const char *path = NULL,
                                    bool force_input = false);

    rt_errno connect(bool  async_connect  = false,
                     const char *version  = FIRE_RT_VERSION_CHECK,
                     const char *platform = FIRE_RT_PLATFORM_CHECK);
    rt_errno disconnect(bool keep_registration = false);
    bool     is_connected() const;
    const char *get_connected_host() const;

    rt_errno workstation_connect(const char *user, const char *passwd = "",
                                 bool async_connect = false);
    
    // This function sets the notify function and closure.  It returns the
    // old notify function and optionally passes back the old closure
    RT_Consumer::NtfyFxn set_notify_function(RT_Consumer::NtfyFxn notify_fxn,
                              void *closure,
                              RT_Consumer::NtfyFxn thread_create_fxn = NULL,
                              RT_Consumer::NtfyFxn thread_destroy_fxn = NULL,
                              void **old_closure = NULL);
    rt_errno get_notify_socket(void *notify_socket) const;
    rt_errno create_mswin_dispatch_window();
    rt_errno destroy_mswin_dispatch_window();
    rt_errno dispatch(int max_wait_time_in_ms = 0);

    rt_errno request(const class RT_Request &req, RT_Consumer::MesgCB cb, 
                     void *closure, int *tag);
    rt_errno request(const class RT_Request &req, RT_Consumer::MesgCB cb, 
                     void *closure, std::vector<int> &tags);
    rt_errno cancel(int tag, void **closure = NULL);

    // For applications who really only need a single consumer
    static RT_Consumer &get_default();

    rt_errno lock()   const;
    rt_errno unlock() const;

    // Some advanced interfaces:
    // get the cached record and puts the pointer in r and returns the closure arg
    // This pointer valid as long as const member functions are being called.
    // Note: dispatch() is not const!
    rt_errno get_record(int tag, const RT_Record **r, void **closure = NULL) const;

    //  gets a copy of the cached record and closure object ptr, thread-safe
    rt_errno get_record(int tag, RT_Record & r, void ** closure = NULL) const;

    // Gets the list of currently available services
    // Should be used after synchronous connects to initialize service lists
    size_t   get_services(class RT_Message *msg) const;

    // updates callback information, and optionally returns the old_closure
    rt_errno update_cb(int tag, RT_Consumer::MesgCB cb, void *closure,
                       void **old_closure = NULL);

    unsigned int num_tags(unsigned int *srv_tags) const;
    int          get_next_tag(int tag_after, unsigned int *srv_tag = NULL) const;

    //  message queue performance statistics
    size_t message_queue_size() const;
    //  average time spent by a message in the internal queue, in microseconds
    double avg_msg_age_in_queue() const;
    //  total number of messages dispatched to user's call-back function(s)
    uint64_t msg_count() const;
    //  get/set threshold to monitor time in cb, set 0 to disable monitoring, microseconds
    size_t cb_warn_time_threshold() const;
    void cb_warn_time_threshold( size_t monitor_threshold_ms );
    //  average time spent inside user's call-back function(s), microseconds
    double avg_cb_time() const;
    //  total number of user call-back function calls
    uint64_t cb_count() const;

    //  plug-ins
    bool enable( RT_Feature &ftr );
    bool disable( RT_Feature &ftr );
    bool is_active( RT_Feature &ftr ) const;

	// setters for enabling unchanged fields
	void set_send_unchanged_fields(bool enable);
	bool get_send_unchanged_fields();

  private:
    RT_Consumer(const RT_Consumer &not_copyable);
    RT_Consumer &operator=(const RT_Consumer &not_copyable);
  public:
    friend class RT_FeatureImpl;
    struct RT_ConsumerImpl *impl;
};

} // namespace

#endif
