#ifndef __RT_LIB_RT_REQUEST_H__
#define __RT_LIB_RT_REQUEST_H__

#include <ostream>
#include <vector>
#include <fire/private/prefix.h>

#ifndef RT_REQ_VERSION
#define RT_REQ_VERSION 0x00000001
#endif

namespace FIRE_NAMESPACE {

class RT_API RT_Request
{
  public:

  enum ReqType {
    REQUEST_TYPE_WATCH  = 1,
    REQUEST_TYPE_SNAP   = 2,
    REQUEST_TYPE_CANCEL = 3,
    REQUEST_TYPE_LOGIN  = 4,
    REQUEST_TYPE_LOGOUT = 5
  };

  enum SymbolType {
    SYMBOL_TYPE_UNKNOWN = 0,
    SYMBOL_TYPE_NATIVE  = 1,
    SYMBOL_TYPE_CUSIP   = 2,
    SYMBOL_TYPE_ISIN    = 3,
    SYMBOL_TYPE_SEDOL   = 4
  };

  RT_Request();
 ~RT_Request();
  RT_Request(const RT_Request &other);
  RT_Request &operator=(const RT_Request &other);

  // Takes null-terminated character strings
  RT_Request(const char *service, const char *key,
             bool snapshot_only = false,
             RT_Request::SymbolType symbol_type = SYMBOL_TYPE_UNKNOWN,
             const char *options = NULL, const char *auth_token = NULL);

  // Takes null-terminated character strings
  RT_Request& set_service    (const char *service);
  RT_Request& set_key        (const char *key);
  RT_Request& set_options    (const char *options);
  RT_Request& set_auth       (const char *auth_token);
  RT_Request& set_snapshot   (bool snapshot_only);
  RT_Request& set_type       (ReqType type);
  RT_Request& set_symbol_type(SymbolType type);


  // set functions with param length (len should not include null-termination)
  RT_Request& set_service    (const char *service,    int len);
  RT_Request& set_key        (const char *key,        int len);
  RT_Request& set_options    (const char *options,    int len);
  RT_Request& set_auth       (const char *auth_token, int len);

  // Returns null-terminated character strings, len is an optional out param
  const char *get_service(int *len = NULL) const;
  const char *get_key    (int *len = NULL) const;
  const char *get_options(int *len = NULL) const;
  const char *get_auth   (int *len = NULL) const;

  bool  is_snapshot()                      const;
  ReqType  get_type()                      const;
  SymbolType  get_symbol_type()            const;
  bool is_extended_req()                   const;

  //
  //private: clients should not use
  //
  RT_Request(ReqType type, const char *service, const char *key,
             const char *options, const char *auth_token);
  RT_Request(const void *src, int src_len, int *error);
  //
  int               get_tag() const;
  unsigned int		  size()    const;
  int         		  serialize(void *dest, unsigned int *bytes_written) const;
  int         		  deserialize(const void *src, unsigned int src_len);
  int         		  deserialize_ext(const void *src, unsigned int src_len, unsigned int ext_version);
  int         		  decode_key(std::vector<std::string> &symbols) const;
  int         		  encode_key(const std::vector<std::string> &symbols);
  // Split a bulk subscription request into multiple single symbol requests.
  int               split(const std::vector<std::string> &req_keys, std::vector<RT_Request> &reqs) const;
  std::vector<int>& tags() const;
  struct RT_RequestRep *rep;
};

//
// The rest is implementation
//
inline bool RT_Request::is_snapshot() const
{  return (get_type() == REQUEST_TYPE_SNAP); }

inline RT_Request& RT_Request::set_snapshot(bool snapshot_only)
{  return set_type(snapshot_only ? REQUEST_TYPE_SNAP : REQUEST_TYPE_WATCH); }

static inline std::ostream& operator<<(std::ostream &os, const RT_Request &req)
{
  int   asz, osz;
  const char *options, *auth;
  options = req.get_options(&osz);
  auth    = req.get_auth(&asz);
  switch (req.get_type()) {
     case RT_Request::REQUEST_TYPE_WATCH:
     case RT_Request::REQUEST_TYPE_SNAP:
       os << "REQ(T:" << req.get_tag() << " K:" << req.get_key(NULL) << " S:" <<
              req.get_service(NULL) << " snap:" << req.is_snapshot();
       if (osz) os << " o:" << options;
       if (asz) os << " a:" << auth;
       os << ")";
       break;
     case RT_Request::REQUEST_TYPE_CANCEL:
       os << "CXL(" << req.get_tag() << ")";
       break;
     case RT_Request::REQUEST_TYPE_LOGOUT:
       os << "LOGOUT(" << req.get_service(NULL) << "," << req.get_key(NULL) << ","
          << auth << "," << options << ")";
       break;
     case RT_Request::REQUEST_TYPE_LOGIN:
       os << "LOGIN(" << req.get_service(NULL) << "," 
          << req.get_key(NULL) << ","
          << "AUTH:{OPAQUE sz=" << asz << "}," << options << ")";
       break;
     default:
       os << "UNKNOWN(" << req.get_service(NULL) << "," << req.get_key(NULL) << ","
          << auth << "," << options << ")";
  }
  return os;
}

} // namespace

#endif
