// include all system header files
#include <iostream>
#include <string>
#include <assert.h>
#ifdef WIN32
#include <winsock2.h>       // used for select()
#include <windows.h>
#else
#include <sys/select.h>
#endif

#include "FDS/rt_api.h"     // include the API header files
#include "FDS/rt_fields.h"  // include the API fields file

using namespace std;    // for convenience
using namespace FDS;    // for convenience

void on_message(int tag, const RT_Message *msg,
                 const RT_Record *rec, void *closure)
{
  // Uncomment to see the message and cached record
  // cout << "Message: " << *msg << endl;
  // cout << "Record: " << *rec << endl;

  // if the server closed the stream close our side as well
  if (msg->is_closed()) { FDF::cancel(tag); }

  if (msg->is_error()) {
    cout << "Error: " << msg->get_error_description() << endl;
    assert(msg->is_closed());
    return;
  }

  // Get message type from message
  string msg_type = msg->get(FIDS::MSG_TYPE).to_string();

  // Get bid and ask from record
  string bid      = rec->get(FIDS::BID_1).to_string();
  string ask      = rec->get(FIDS::ASK_1).to_string();

  cout << "Update: " << msg_type << " Bid: " << bid 
       << " Ask: " << ask << endl;
}

void on_static_message(int tag, const RT_Message *msg,
                        const RT_Record *rec, void *closure)
{
  cout << "Message: " << *msg << endl;
  cout << "Record: " << *rec << endl;

  // No reason to check is->closed(), since we only want a single
  // message.  So just call cancel() after processing.  BTW, Since
  // the app is asking for a snapshot, the message will be closed
  // anyway. i.e. (msg->is_closed() == true).
  // 
  FDF::cancel(tag);
}

int main(int argc, char **argv)
{
  rt_errno err;
  
  if (argc <= 2) {
     RT_LOG_RAW("Usage: " << argv[0] << " <username>:<password>@<hostname>\n");
     RT_LOG_RAW("Example: " << argv[0] << " user:passwd@api-stage.df.factset.com\n");
     return 3;
  }
  string connection(argv[1]);

  RT_FieldMap::create("../../../etc/rt_fields.xml");
  // if using basic authentication
  FDF::set_connection_info(connection.c_str());
  // if using OTP comment the line above and uncomment the line below with correct details
  // FDF::set_connection_info("api-stage.df.factset.com", "client","AAAA", NULL,
  // NULL, "C:\\Path\\To\\Counterfile", false);

  err = FDF::connect();
  if (err) { cerr << "conn: " << err << endl; return (int)err; }

  int tag;
  // create a real-time RT_Request for service=FDS1, symbol=FDS-USA
  RT_Request req("FDS1", "FDS-USA");
  FDF::request(req, on_message, NULL, &tag);
  cout << "made a request for " << req << " tag=" << tag << endl;

  // create a static RT_Request for service=FDS1, symbol=IBM-USA
  // The parameter "true" sets the snapshot_only flag
  int stag;
  RT_Request sreq("FDS1", "IBM-USA", true);
  FDF::request(sreq, on_static_message, NULL, &stag);
  cout << "made a static request for " << sreq << " tag=" << stag << endl;

  // dispatch messages
  while (true) {
    FDF::dispatch(-1);
  }

  /*
   * Integration with a select loop
   *
  int api_fd;
  int select_max = 0;
  fd_set read_fds;

  err = FDF::get_notify_socket(&api_fd);
  if (err) { cerr << "No resource: " << endl; return (int)err; }

  // need to reset the select max descriptor
  if (api_fd > select_max) select_max = api_fd;

  while (true) {
    FD_ZERO(&read_fds);
    //FD_SET(...);                // set other app descriptors
    FD_SET(api_fd, &read_fds);  // monitor the API socket for read

    int ndesc = select(select_max + 1, &read_fds, NULL, NULL, NULL);
    if (ndesc < 0) continue;    // handle errors

    // check the API socket descriptor
    if (ndesc && FD_ISSET(api_fd, &read_fds))
      // the API wants to dispatch messages, call it with a 0 timeout
      FDF::dispatch(0);

      // handle other descriptors
  }
  */

  return 0;
}


